#version 430



layout(location = 1) in vec3 INPUT_POSITION;
layout(location = 6) in vec3 INPUT_NORMAL;
layout(location = 9) in vec2 uvs;


layout(location = 2) uniform vec3 COLOR;

//camera position
layout(location = 10) uniform vec3 eye;
//((colx, coly, colz), (dirx, diry, dirz))
layout(location = 11) uniform mat2x3 directionalLight;
//((colx, coly, colz), (posx, posy, posz), (AConst, ALin, AQuadr)
layout(location = 12) uniform mat3 pointLight1;
//((colx, coly, colz), (posx, posy, posz), , (AConst, ALin, AQuadr), (dirx, diry, dirz)
layout(location = 13) uniform mat4x3 frontLight1;
layout(location = 14) uniform mat4x3 frontLight2;

layout(location = 20) uniform int textureSet;
layout(location = 21)uniform sampler2D screenTexture;
layout(location = 22) uniform samplerCube envMap;

layout(location = 31)uniform float metallic;
layout(location = 32)uniform float roughness;
layout(location = 33)uniform float ao;
layout(location = 34)uniform float reflectionFactor;


layout(location = 0) out vec4 fragColor;



float PI = 3.14159265359f;

//Trowbridge-Reitz GGX normal distribution function
//approximating number of microfacets aligned to the halfway Vector
float DistributionGGX(vec3 normal, vec3 halfway, float roughness)
{
	float a      = roughness*roughness;
    float a2     = a*a;
    float NdotH  = max(dot(normal, halfway), 0.0);
    float NdotH2 = NdotH*NdotH;
	
    float num   = a2;
    float denom = (NdotH2 * (a2 - 1.0) + 1.0);
    denom = PI * denom * denom;
	
    return num / denom;
}


//approximating microfacet shadowing and occlusion
float GeometrySchlickGGX(float NdotV, float roughness)
{
    float r = (roughness + 1.0);
    float k = (r*r) / 8.0;

    float num   = NdotV;
    float denom = NdotV * (1.0 - k) + k;
	
    return num / denom;
}
  
float GeometrySmith(vec3 N, vec3 V, vec3 L, float roughness)
{
   float NdotV = max(dot(N, V), 0.0);
    float NdotL = max(dot(N, L), 0.0);
    float ggx2  = GeometrySchlickGGX(NdotV, roughness);
    float ggx1  = GeometrySchlickGGX(NdotL, roughness);
	
    return ggx1 * ggx2;
}

//With cosTheta being the dot product result between the surface's normal n and the halfway h direction. 
//calculates how much light is reflected and how much is refracted
vec3 fresnelSchlick(float cosTheta, vec3 F0)
{
    return F0 + (1.0 - F0) * pow(1.0 - cosTheta, 5.0);
}



void main(){
	
	vec3 sunDir = -normalize(directionalLight[1]);
	vec3 sunCol = directionalLight[0];

	vec3 pointDir = normalize( pointLight1[1] - INPUT_POSITION);
	vec3 pointCol = pointLight1[0];
	float dist = length(pointLight1[1]-INPUT_POSITION);
	float attenuationPoint = 1 / (pointLight1[2].x+dist*pointLight1[2].y+dist*dist*pointLight1[2].z);

	vec3 normal = normalize(INPUT_NORMAL);
	vec3 viewDirection = normalize(eye-INPUT_POSITION);
	vec3 viewDirectionReflection = reflect(-viewDirection, normal);

	vec3 myTexture;
	vec3 reflectColor;

	//set base color
	if(textureSet==1){
		myTexture = vec3(texture2D(screenTexture, uvs))+COLOR;		
	}
	else{
		myTexture = COLOR;	
	}
	//add reflection
	reflectColor = vec3(texture(envMap, viewDirectionReflection));
	myTexture = (1-reflectionFactor) * myTexture + (reflectColor*reflectionFactor);

	//setup light sources in array
	const int numberOfLights = 4;

	vec3 lightColors[numberOfLights]; 
	vec3 lightDirections[numberOfLights]; 
	float attenuations[numberOfLights];

	lightColors[0] = sunCol;
	lightDirections[0] = sunDir;
	attenuations[0] = 1.f;


	lightDirections[1] = pointDir;
	lightColors[1] = pointCol;
	attenuations[1] = attenuationPoint;


	float thetaFront1 = dot(normalize(frontLight1[1] - INPUT_POSITION), normalize(-frontLight1[3]));
	if(thetaFront1 > 0.9f) 
	{       
		lightColors[2] = frontLight1[0];
	}else{
		lightColors[2] = vec3(0,0,0);
	}
	lightDirections[2] =  frontLight1[1] - INPUT_POSITION;
	attenuations[2] = 1 / (frontLight1[2].x+dist*frontLight1[2].y+dist*dist*frontLight1[2].z);


	float thetaFront2 = dot(normalize(frontLight2[1] - INPUT_POSITION), normalize(-frontLight2[3]));
	if(thetaFront2 > 0.9f) 
	{       
		lightColors[3] = frontLight2[0];
	}else{
		lightColors[3] = vec3(0,0,0);
	}
	lightDirections[3] =  frontLight2[1] - INPUT_POSITION;
	attenuations[3] = 1 / (frontLight2[2].x+dist*frontLight2[2].y+dist*dist*frontLight2[2].z);
	
	
	vec3 ilumination = vec3(0,0,0);

	for (int i = 0; i < numberOfLights; i++){

		//direct light
		vec3 halfway = normalize(viewDirection + lightDirections[i]);
		vec3 radiance     = lightColors[i] * attenuations[i]; 

		
		//calculate how much ligh gets reflected and how much gets refracted
		vec3 F0 = vec3(0.04);					//surface reflection at zero incidence 
		F0 = mix(F0, myTexture, metallic);		//for metalic textures the value is tinted for other materials 0.04 is a good aprox
		vec3 F  = fresnelSchlick(max(dot(halfway, viewDirection), 0.0), F0);
		
		
		//statistic aproximation of microfacets aligned to halfway vector 
		float NDF = DistributionGGX(normal, halfway, roughness);    
		
		//statistic apoximation of microfacets overshadowing each other
		float G   = GeometrySmith(normal, viewDirection, lightDirections[i], roughness);       

		vec3 numerator    = NDF * G * F;
		float denominator = 4.0 * max(dot(normal, viewDirection), 1.0) * max(dot(normal, lightDirections[i]), 1.0);
		vec3 specular     = numerator / max(denominator, 0.001); 



		vec3 kS = F;					//energy of light that gets reflected
		vec3 kD = vec3(1.0) - kS;		//energy of light that gets refrected
  
		kD *= (1.0 - metallic);	

		float NdotL = max(dot(normal, lightDirections[i]), 0.0);        
		ilumination = ilumination + (kD * myTexture / PI   + specular) * radiance * NdotL;
	}

	vec3 ambient = vec3(0.1) * myTexture * ao;
	vec3 color   = ambient + ilumination; 

	//gamma correction
	color = color / (color + vec3(1.0));
	color = pow(color, vec3(1.0/2.2)); 
   
    fragColor = vec4(color, 1.0);
   }

